import pygame

import scene_handler
from better_math import Vector2i
from elements.py_button import Button, ToggleButton
from pygame_scene import PyGameScene
from elements.text import Text, TextEdit, UnderlinedTextEdit
from elements.texture import Texture, StagedTexture


class PreviewScene(PyGameScene):
    def update(self):
        super().update()
        # Adding a Text Object
        self.drawables.append(Text("This is a Text","arial",pygame.Rect(scene_handler.camera_size.x // 2,60,0,40)))


        # Specifying the Size and Position of a Button
        button_rect = pygame.Rect(0,120,250,50)
        button_rect.centerx = scene_handler.camera_size.x // 2


        # Adding a Button Object
        self.drawables.append(Button("This is a Button",(scene_handler.camera_size.x // 2, 130),(250,40),pygame.font.SysFont("arial",16),lambda: print("You Clicked A Button")))


        # Adding a Toggle Button Object, DISCLAIMER: This is buggy and doesn't work as Supposed.
        self.drawables.append(ToggleButton("Test",(scene_handler.camera_size.x // 2, 130),(0,0),pygame.font.SysFont("arial",16),lambda new_value: print("You Toggled A Button"),text_color=(0,0,0)))


        # Adding a Input Field
        self.drawables.append(TextEdit(
            "", # Here Goes the Text that is Shown
            "Enter Text:", # Here Goes a Placeholder, a Gray text that's shown until your write text of focus
            "arial", # The Name of the Font you want to use
            pygame.Rect(scene_handler.camera_size.x // 2 - 125,200,250,40) # The Position and Size on the Screen
        ))


        # Adding a Underlined Input Field, this is handy if you want to make smthn like Hangman
        self.drawables.append(UnderlinedTextEdit(
            "", # The Text Shown on the Lines
            6, # The Amount of Lines Shown
            "arial", # The Name of the Font you want to use
            Vector2i(scene_handler.camera_size.x // 2,260), # The Position and Size on the Screen
            centered=True, # If the object is supposed to be Centered
            skip=[2] # The Indexes of Lines to be Skipped
        ))


        # Adding a texture, this is for Rendering simple Textures, it's not supposed to be used as GameObjects as it is supposed to only be changed in the update method
        self.drawables.append(Texture("textures/test.png",pygame.Rect(0,0,128,128),texture_size=Vector2i(16,16)))

        # Adding a Staged Texture, in My Case in the Render Method i am Changing the stage every 90 Frames/1.5 seconds on 60 FPS
        self.staged_texture = StagedTexture(["textures/hangman_0.png","textures/hangman_1.png","textures/hangman_2.png","textures/hangman_3.png","textures/hangman_4.png","textures/hangman_5.png","textures/hangman_6.png","textures/hangman_7.png","textures/hangman_8.png"],pygame.Rect(0,128,512,512),texture_size=Vector2i(128,128))
        self.drawables.append(self.staged_texture)


    frame_amount = 0


    def render(self, screen, events) -> bool:
        pygame.draw.rect(screen,(255,255,255),(0,0,scene_handler.camera_size.x,scene_handler.camera_size.y))
        self.frame_amount += 1
        self.staged_texture.stage = (self.frame_amount // 90) % 10
        super().render(screen,events)
        return True


import pygame

import math
from better_math import Vector2i


class DrawContext:
    def __init__(self, screen: pygame.Surface, position_offset: Vector2i = Vector2i(0, 0)):
        self.screen = screen
        self.position_offset = position_offset
        self._rendered_texture_cache : dict[tuple[tuple[int, int, int, int],Vector2i,Vector2i,tuple[int,int,int]], pygame.Surface] = {}
    _texture_cache: dict[str, pygame.Surface] = {}
    def _load_texture(self, path: str) -> pygame.Surface:
        if path not in self._texture_cache:
            self._texture_cache[path] = pygame.image.load(path).convert_alpha()
        return self._texture_cache[path]

    def texture(
            self,
            texture_path: str,
            pos: Vector2i,
            uv: tuple[int, int, int, int] = (0, 0, -1, -1),
            size: Vector2i = Vector2i(32, 32),
            texture_size: Vector2i = Vector2i(16, 16),
            color: tuple[int, int, int] = (255, 255, 255),
    ) -> None:
        """
        Schnelle Version:
        - konvertierte Texturen & (tw,th)-Skalierung werden gecacht
        - Farbmultiplikation wird gecacht
        - Tiling via Surface.blits (Batch) statt Python-Doppelschleife
        - überspringt Arbeit in häufigen Trivialfällen
        """
        position = (int(pos.x + self.position_offset.x), int(pos.y + self.position_offset.y))
        if (uv, size, texture_size, color) in self._rendered_texture_cache:
            self.screen.blit(self._rendered_texture_cache[(uv, size, texture_size, color)], position)

        # --- kleine Helfer & Caches ---------------------------------------------
        # erzeuge Caches einmal z.B. im __init__: self._tile_cache = {}, self._color_cache = {}
        # self._load_texture(texture_path) sollte bereits convert_alpha() benutzen!
        def _get_scaled_tile(img, tw, th):
            key = (id(img), tw, th)
            surf = img if (img.get_width() == tw and img.get_height() == th) \
                else pygame.transform.scale(img, (tw, th))
            # wichtig: gleiche Pixel-Format für schnelle Blits
            if surf.get_bitsize() != self.screen.get_bitsize() or not surf.get_masks()[3]:
                surf = surf.convert_alpha()
            return surf

        def _get_colored_tile(tile, color):
            if color == (255, 255, 255):
                return tile
            key = (id(tile), color)
            surf = tile.copy()
            surf.fill((*color, 255), special_flags=pygame.BLEND_RGBA_MULT)
            return surf

        # ------------------------------------------------------------------------
        img = self._load_texture(texture_path)  # sollte convert_alpha liefern

        tw, th = int(texture_size.x), int(texture_size.y)
        if tw <= 0 or th <= 0:
            return

        u0, v0, uW, vH = uv
        if uW == -1: uW = tw
        if vH == -1: vH = th
        if uW <= 0 or vH <= 0:
            return

        # 1) Tile auf (tw,th) skalieren (aus Cache)
        base_tile = _get_scaled_tile(img, tw, th)

        # 2) Farbe anwenden (aus Cache)
        tile = _get_colored_tile(base_tile, color)

        # Fast-Path A: genau ein Tile, kein Offset → direkte Verwendung
        if (uW == tw and vH == th) and (u0 % tw == 0) and (v0 % th == 0):
            # ggf. auf Zielgröße skalieren
            if size.x != uW or size.y != vH:
                render_surface = pygame.transform.scale(tile, (int(size.x), int(size.y)))
                self.screen.blit(render_surface, position)
            else:
                self.screen.blit(tile, position)
            return

        # 3) Musterfläche (uW x vH) füllen (mit Offset) – per Batch-Blit
        pattern_surface = pygame.Surface((uW, vH), flags=pygame.SRCALPHA)

        # Startposition so wählen, dass das erste sichtbare Tile links/oben außerhalb beginnt
        offset_x = (-u0) % tw
        offset_y = (-v0) % th
        start_x = offset_x - tw
        start_y = offset_y - th

        tiles_x = (uW // tw) + 3
        tiles_y = (vH // th) + 3

        # Alle Zielpositionen vorab berechnen und in einem Rutsch blitten
        dests = []
        bx = start_x
        for ix in range(tiles_x):
            by = start_y
            for iy in range(tiles_y):
                dests.append((bx, by))
                by += th
            bx += tw
        # Batch-Blit (deutlich weniger Python-Overhead als Doppel-Schleife)
        pattern_surface.blits([(tile, d) for d in dests])

        # 4) Auf Zielgröße bringen (Nearest Neighbor)
        if size.x != uW or size.y != vH:
            render_surface = pygame.transform.scale(pattern_surface, (int(size.x), int(size.y)))
        else:
            render_surface = pattern_surface

        # 5) Zeichnen
        self._rendered_texture_cache[(uv, size, texture_size, color)] = render_surface
        self.screen.blit(render_surface, position)



    def text(self, text: str, font: str, pos: Vector2i, size: int, color=(0, 0, 0)):
        """Zeichnet Text ab (pos.x, pos.y) mit pygame.font.SysFont(font, size)."""
        target = self.screen  # erwartet: pygame.Surface vorhanden
        fnt = pygame.font.SysFont(font, size)  # pygame.font ist bereits initialisiert

        x, y = int(pos.x), int(pos.y)
        lines = text.split("\n")
        line_h = fnt.get_linesize()

        for line in lines:
            if line == "":
                y += line_h
                continue
            surf = fnt.render(line, True, color)

            target.blit(surf, (x, y))
            y += line_h




import pygame

import draw_context
import pygame_scene
from draw_context import DrawContext
from better_math import Vector2i
from pygame_scene import PyGameUIHandler


class Text(PyGameUIHandler):
    def __init__(self, text: str,font, rect: pygame.Rect):
        self.text = text
        self.rect = rect
        self.font = font
        centerx = rect.centerx
        rect.w = pygame.font.SysFont(font,rect.h).size(text)[0]
        rect.centerx = centerx

        super().__init__()

    def get_rect(self) -> pygame.Rect:
        return self.rect

    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        context = DrawContext(surface)
        context.text(self.text,self.font,Vector2i(self.get_rect().x,self.get_rect().y),self.get_rect().h)
        super().draw(surface,offset)

class TextEdit(PyGameUIHandler):
    def __init__(self, text: str, placeholder: str, font, rect: pygame.Rect, can_write = None):
        self.text = text
        self.placeholder = placeholder
        self.font = font
        self.rect = rect
        self.can_write = can_write
        super().__init__()

    hovered = False
    animation_tick = 0
    def get_rect(self) -> pygame.Rect:
        return self.rect
    def handle_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            if self.get_rect().collidepoint(event.pos[0],event.pos[1]):
                self.hovered = True
                return True
            else:
                self.hovered = False
        elif event.type == pygame.MOUSEBUTTONDOWN:
            if self.hovered:
                if pygame_scene.focused_ui_handler is not None:
                    pygame_scene.focused_ui_handler.no_longer_focused()
                pygame.key.set_repeat(300, 30)
                pygame_scene.focused_ui_handler = self
                return True
        elif event.type == pygame.TEXTINPUT and self.is_focused():
            if self.can_write is not None:
                if not self.can_write(self.text + event.text):
                    return True
            self.text += event.text
            return True
        elif event.type == pygame.KEYDOWN and self.is_focused():
            if event.key == pygame.K_BACKSPACE:
                self.text = self.text[:-1]
                return True
        return False
    def no_longer_focused(self):
        print("No Longer Focused")
        pygame.key.set_repeat(0)
    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        bg_texture = (170,170,170)
        if self.hovered:
            bg_texture = (190,190,190)
        pygame.draw.rect(surface,bg_texture,self.get_rect(),border_radius=10)
        pygame.draw.rect(surface,(70,70,70),self.get_rect(),width=2,border_radius=10)
        context = draw_context.DrawContext(surface,offset)
        text = self.text
        if self.is_focused():
            if (self.animation_tick % 50) < 25:
                text += "|"
            self.animation_tick += 1
        else:
            self.animation_tick = 0
        if text == "" and not self.is_focused():
            context.text(self.placeholder,self.font,Vector2i(self.get_rect().x + 5, self.get_rect().center[1] - 8),16,(120,120,120))
        else:
            context.text(text,self.font, Vector2i(self.get_rect().x + 5, self.get_rect().center[1] - 8), 16,
                         (0,0,0))

class UnderlinedTextEdit(TextEdit):
    def __init__(self, text: str, line_amount: int, font, pos : Vector2i, can_write = None, centered : bool = False, skip : list[int] = []):
        self.line_amount = line_amount
        self.can_write_underlined = can_write
        self.skip = skip
        def check_for_lenght(new_text):
            print(len(new_text) > line_amount and not (can_write is not None and can_write(new_text)))
            if len(new_text) > line_amount and not (can_write is not None and can_write(new_text)):
                return False,
            return True
        rect = pygame.Rect(pos.x,pos.y,50 * line_amount - 10,40)
        if centered:
            rect.centerx = pos.x
        super().__init__(text, "", font, rect,check_for_lenght)

    def handle_event(self, event):
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_BACKSPACE:
                if self.text.endswith(" "):
                    self.text = self.text[:-1]
        val = super().handle_event(event)
        if event.type == pygame.TEXTINPUT:
            if len(self.text) in self.skip:
                self.text += " "
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_BACKSPACE:
                if self.text.endswith(" "):
                    self.text = self.text[:-1]
        return val

    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        x = 0
        context = draw_context.DrawContext(surface,offset)
        for line in range(self.line_amount):
            if line in self.skip:
                x += 50
                continue
            pygame.draw.rect(surface,(0,0,0),(x + self.get_rect().x,self.get_rect().y + self.get_rect().height - 6,40,6))
            if len(self.text) > line:
                context.text(self.text[line],self.font,Vector2i(x + self.get_rect().x + (20 - pygame.font.SysFont(self.font,40).size(self.text[line])[0] // 2),self.get_rect().y),40)
            x += 50


import pygame

from draw_context import DrawContext
import pygame_scene
from better_math import Vector2i


class Button(pygame_scene.PyGameUIHandler):
    def __init__(self, text, center, size, font, on_click, color=(60, 120, 255), hover_color=(40, 90, 210),
                 text_color=(240, 240, 240), on_click_values = ()):
        super().__init__()
        self.collision_rect = pygame.Rect(center[0], center[1], size[0], size[1])
        self.text = text
        self.rect = pygame.Rect(0, 0, *size)
        self.rect.center = center
        self.font = font
        self.on_click = on_click
        self.color = color
        self.hover_color = hover_color
        self.text_color = text_color
        self.hovered = False
        self.pressed = False
        self.on_click_values = on_click_values
        if not (font is None):
            self.text_surf = self.font.render(self.text, True, self.text_color)
            self.text_rect = self.text_surf.get_rect(center=self.rect.center)
        else:
            self.text_surf = None


    def handle_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            self.hovered = self.get_rect().collidepoint(event.pos)
            return True
        elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            if self.hovered:
                self.pressed = True
                return True
        elif event.type == pygame.MOUSEBUTTONUP and event.button == 1:
            if self.pressed and self.hovered:
                self.on_click(*self.on_click_values)
                return True
            self.pressed = False
        return False

    def get_rect(self) -> pygame.Rect:
        return self.rect.move(self.parent.get_rect().x, self.parent.get_rect().y)

    def draw(self, surface,offset : Vector2i = Vector2i(0, 0)):
        col = self.hover_color if self.hovered else self.color
        pygame.draw.rect(surface, col, self.rect.move(offset.x,offset.y), border_radius=10)
        pygame.draw.rect(surface, (0, 0, 0), self.rect.move(offset.x,offset.y), 2, border_radius=10)
        self.text_rect.center = self.rect.center
        surface.blit(self.text_surf, self.text_rect.move(offset.x,offset.y))


class ToggleButton(Button):
    toggled = False

    def _do_nothing(self, new_value):
        pass

    def __init__(self, text, center, size, font, on_changed=_do_nothing, color=(60, 120, 255), hover_color=(40, 90, 210),
                 text_color=(240, 240, 240),  on_click_values = ()):
        def toggle():
            self.toggled = not self.toggled
            on_changed(self.toggled)

        self.on_changed = on_changed
        super().__init__(text, center, size, font, toggle,color,hover_color,text_color,on_click_values=on_click_values)

    def get_rect(self) -> pygame.Rect:
        button_rect = self.rect.move(60, (self.rect.y // 2) + 7)
        button_rect.w = 40
        button_rect.h = 20
        return button_rect.move(self.parent.get_rect().x, self.parent.get_rect().y)

    def draw(self, surface,offset : Vector2i = Vector2i(0, 0)):
        col = self.hover_color if self.hovered else self.color
        button_rect = self.rect.move(60, (self.rect.y // 2) + 7)
        button_rect.w = 40
        button_rect.h = 20
        pygame.draw.rect(surface, (60, 60, 60), button_rect.move(offset.x,offset.y), border_radius=10)
        pygame.draw.rect(surface, (0, 0, 0), button_rect.move(offset.x,offset.y), 2, border_radius=10)
        toggle_pos = -20
        if self.toggled:
            toggle_pos = 20
        circle_pos = (button_rect.center[0] + toggle_pos + offset.x, button_rect.center[1] + offset.y)
        pygame.draw.circle(surface, col, circle_pos, 12)
        pygame.draw.circle(surface, (0, 0, 0), circle_pos, 12, 2)
        self.text_rect.center = (button_rect.center[0] - (self.font.size(self.text)[0] // 2), button_rect.center[1])

        surface.blit(self.text_surf, self.text_rect.move(-(self.text_rect.w // 2) - 20, 0))
        pygame_scene.PyGameUIHandler.draw(self, surface)

class TextureButton(Button):
    def __init__(self, texture_path, center, size, on_click,texture_size=Vector2i(32,32), on_click_values=()):
        self.texture_path = texture_path
        self.texture_size = texture_size
        super().__init__("", center, size, None, on_click, on_click_values=on_click_values)

    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        draw_context = DrawContext(surface,offset)
        draw_context.texture(self.texture_path, Vector2i(self.rect.x, self.rect.y), size=Vector2i(self.get_rect().w, self.get_rect().h),texture_size=self.texture_size)


import pygame
from pygame import MOUSEBUTTONDOWN

from better_math import Vector2i
import scene_handler

class PyGameUIHandler:

    def is_focused(self):
        return self == focused_ui_handler
    def __init__(self):
        self.visible = True
        self._children = []
        self.parent = EmptyUIHandler()

    def add_child(self, child):
        self._children.append(child)
        child.parent = self

    text = ""


    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        for child in self._children:
            if child.visible:
                child.draw(surface,Vector2i(self.get_rect().x, self.get_rect().y) + offset)

    def handle_event(self, event) -> bool:
        for child in self._children:
            if child.visible:
                if child.handle_event(event):
                    return True
        return False
    def get_rect(self) -> pygame.Rect:
        return pygame.Rect(0, 0, 0, 0).move(self.parent.get_rect().x, self.parent.get_rect().y)

    def collidepoint(self, x, y):
        for drawable in self._children:
            if not drawable.visible:
                continue
            if drawable.get_rect().collidepoint(x, y):
                return True
            elif drawable.collidepoint(x, y):
                return True
        else:
            return False

    def no_longer_focused(self):
        pass


focused_ui_handler : PyGameUIHandler = None
class EmptyUIHandler(PyGameUIHandler):
    def __init__(self):
        pass

    def get_rect(self) -> pygame.Rect:
        return pygame.Rect(0, 0, 0, 0)
class PyOverlay(PyGameUIHandler):
    def __init__(self, rect: pygame.Rect, color: tuple[int, int, int] = (0, 0, 0)):
        super().__init__()
        self.rect = rect
        self.color = color

    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        pygame.draw.rect(surface, self.color, self.rect)
        super().draw(surface, offset)
    def get_rect(self) -> pygame.Rect:
        return self.rect.move(self.parent.get_rect().x, self.parent.get_rect().y)


class PyGameScene:
    drawables: list[PyGameUIHandler] = []

    def __init__(self):
        if pygame.get_init():
            self.update()

    def render(self, screen, events) -> bool:
        for event in events:
            if event.type == pygame.QUIT:
                return False
            is_handled = False
            for drawable in self.drawables:
                if drawable.visible:
                    if drawable.handle_event(event):
                        is_handled = True
            if not is_handled and event.type == pygame.MOUSEBUTTONDOWN:
                global focused_ui_handler
                if focused_ui_handler is not None:
                    focused_ui_handler.no_longer_focused()
                focused_ui_handler = None

        for drawable in self.drawables:
            if drawable.visible:
                if drawable.text == "0":
                    print("Drawing Button")
                drawable.draw(screen)
        if scene_handler.debugging:
            for drawable in self.drawables:
                pygame.draw.rect(screen, pygame.Color("red"), drawable.get_rect(),1)
        return True

    def update(self):
        self.drawables.clear()

    def collidepoint(self, x, y):
        for drawable in self.drawables:
            if not drawable.visible:
                continue
            if drawable.get_rect().collidepoint(x, y):
                return True
            elif drawable.collidepoint(x, y):
                return True
        return False





import pygame

import scene_handler
from better_math import Vector2i
from preview_scene import PreviewScene
from pygame_scene import PyGameScene

clock = pygame.time.Clock()

def setup(default_scene : PyGameScene,window_name="PyGameUI Window"):
    pygame.init()
    pygame.font.init()
    pygame.display.set_caption(window_name)
    screen = pygame.display.set_mode((scene_handler.camera_size.x,scene_handler.camera_size.y),pygame.RESIZABLE)
    running = True
    scene_handler.current_scene = default_scene
    default_scene.update()
    while running:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                running = False
            if event.type == pygame.VIDEORESIZE:
                scene_handler.camera_size = Vector2i(event.w,event.h)
                scene_handler.current_scene.update()
        if running:
            running = scene_handler.current_scene.render(screen,events)
        pygame.display.update()
        clock.tick(scene_handler.FPS)


if __name__ == "__main__":
    setup(PreviewScene())





class Vector2i:

    def __init__(self, x, y):
        self.x = x
        self.y = y
    def __add__(self, other):
        if isinstance(other, Vector2i):
            return Vector2i(self.x + other.x, self.y + other.y)
        return Vector2i(self.x + other, self.y + other)

    def __mul__(self, other):
        if isinstance(other, Vector2i):
            return Vector2i(self.x * other.x, self.y * other.y)
        return Vector2i(self.x * other, self.y * other)

    def __truediv__(self, other):
        if isinstance(other, Vector2i):
            return Vector2i(self.x // other.x, self.y // other.y)  # int-Division
        return Vector2i(self.x // other, self.y // other)

    def __abs__(self):
        return Vector2i(abs(self.x), abs(self.y))

    def __neg__(self):
        return Vector2i(-self.x, -self.y)

    def __pow__(self, power, modulo=None):
        if modulo is None:
            return Vector2i(self.x ** power, self.y ** power)
        return Vector2i(pow(self.x, power, modulo), pow(self.y, power, modulo))

    def __repr__(self):
        return f"Vector2i({self.x}, {self.y})"
    def __str__(self):
        return f"({self.x}, {self.y})"



import pygame

import draw_context
from better_math import Vector2i
from pygame_scene import PyGameUIHandler


class Texture(PyGameUIHandler):
    def __init__(self, texture_path, rect : pygame.Rect, uv : tuple[int,int,int,int] = (0,0,-1,-1),texture_size : Vector2i = Vector2i(128,128)):
        self.texture_path = texture_path
        self.rect = rect
        self.uv = uv
        self.texture_size = texture_size
        super().__init__()
    def get_rect(self) -> pygame.Rect:
        return self.rect

    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        context = draw_context.DrawContext(surface)
        context.texture(self.texture_path,Vector2i(self.get_rect().x,self.get_rect().y),self.uv,Vector2i(self.get_rect().w,self.get_rect().h),self.texture_size)

class StagedTexture(PyGameUIHandler):
    def __init__(self,texture_paths : list[str],rect : pygame.Rect,uv : tuple[int,int,int,int] = (0,0,-1,-1),texture_size : Vector2i = Vector2i(128,128)):
        self.texture_paths = texture_paths
        self.rect = rect
        self.uv = uv
        self.texture_size = texture_size
        self.stage = 0
        super().__init__()
    def get_rect(self) -> pygame.Rect:
        return self.rect
    def draw(self, surface, offset: Vector2i = Vector2i(0, 0)):
        context = draw_context.DrawContext(surface)
        for stage in range(self.stage):
            context.texture(self.texture_paths[stage], Vector2i(self.get_rect().x, self.get_rect().y), self.uv,
                            Vector2i(self.get_rect().w, self.get_rect().h), self.texture_size)

